from typing import Any, Dict
from pydantic_resolve.exceptions import GlobalLoaderFieldOverlappedError


def merge_dicts(a: Dict[str, Any], b: Dict[str, Any]):
    overlap = set(a.keys()) & set(b.keys())
    if overlap:
        raise GlobalLoaderFieldOverlappedError(f'loader_params and global_loader_param have duplicated key(s): {",".join(overlap)}')
    else:
        return {**a, **b}


from typing import Any, Dict
from inspect import isfunction
from dataclasses import is_dataclass
from pydantic import BaseModel
import pydantic_resolve.constant as const
from pydantic_resolve.compat import PYDANTIC_V2
from pydantic_resolve.utils.class_util import safe_issubclass, is_pydantic_field_required_field, get_pydantic_field_items


def _get_required_fields(kls: BaseModel):
    required_fields = []

    items = get_pydantic_field_items(kls)

    for fname, field in items:
        if is_pydantic_field_required_field(field):
            required_fields.append(fname)


    # 2. get resolve_ and post_ target fields
    for f in dir(kls):
        if f.startswith(const.RESOLVE_PREFIX):
            if isfunction(getattr(kls, f)):
                required_fields.append(f.replace(const.RESOLVE_PREFIX, ''))

        if f.startswith(const.POST_PREFIX):
            if isfunction(getattr(kls, f)):
                required_fields.append(f.replace(const.POST_PREFIX, ''))

    return required_fields


def model_config_v1(default_required: bool = True):
    """
    - default_required: 
        if resolve field has default value, it will not be listed in schema['required']
        set default_required=True to add it into required list.
    """
    def _get_pydantic_model(kls):
        if safe_issubclass(kls, BaseModel):
            return kls
        elif hasattr(kls, '__pydantic_model__'):
            return kls.__pydantic_model__
        else:
            raise AttributeError(f'target class {kls.__name__} is not subclass of BaseModel, or decorated by pydantic.dataclass')

    def wrapper(kls):
        pydantic_model = _get_pydantic_model(kls)
        def _schema_extra(schema: Dict[str, Any], model) -> None:
            # define schema.properties
            excludes = set()

            if pydantic_model.__exclude_fields__:
                for k in pydantic_model.__exclude_fields__.keys():
                    excludes.add(k)

            props = {}
            for k, v in schema.get('properties', {}).items():
                if k not in excludes:
                    props[k] = v
            schema['properties'] = props

            # define schema.required
            if default_required:
                fnames = _get_required_fields(model)
                schema['required'] = fnames

        pydantic_model.__config__.schema_extra = staticmethod(_schema_extra)
        return kls
    return wrapper


def model_config_v2(default_required: bool=True):
    """
    in pydantic v2, we can not use __exclude_field__ to set hidden field in model_config hidden_field params
    model_config now is just a simple decorator to remove fields (with exclude=True) from schema.properties
    and set schema.required for better schema description. 
    (same like `output` decorator, you can replace output with model_config)

    it keeps the form of model_config(params) in order to extend new features in future

    update:

    in fastapi + pydantic v2, this function will be handled internal automatically with mode: serilization
    you can remove the model_config_v2

    reference: fastapi/_compat.py::get_definitions
    """
    def wrapper(kls):
        if is_dataclass(kls):
            return kls
        elif safe_issubclass(kls, BaseModel):
            # TODO: check the behavior of generating json schema in other frameworks using pydantic
            def build():
                def _schema_extra(schema: Dict[str, Any], model) -> None:
                    # 1. collect exclude fields and then hide in both schema and dump (default action)
                    excluded_fields = [k for k, v in kls.model_fields.items() if v.exclude == True]
                    props = {}

                    # config schema properties
                    for k, v in schema.get('properties', {}).items():
                        if k not in excluded_fields:
                            props[k] = v
                    schema['properties'] = props

                    # config schema required (fields with default values will not be listed in required field)
                    # and the generated typescript models will define it as optional, and is troublesome in use
                    if default_required:
                        fnames = _get_required_fields(model)
                        if excluded_fields:
                            fnames = [n for n in fnames if n not in excluded_fields]
                        schema['required'] = fnames

                return _schema_extra

            kls.model_config['json_schema_extra'] = staticmethod(build())
            return kls
        else:
            raise AttributeError(f'target class {kls.__name__} is not BaseModel or dataclass')
    return wrapper


model_config = model_config_v2 if PYDANTIC_V2 else model_config_v1


import os
import asyncio
import warnings
import contextvars
from inspect import iscoroutine
from typing import TypeVar, Dict, Type, Callable, Any, Optional
from aiodataloader import DataLoader
from types import MappingProxyType

from pydantic_resolve import analysis
from pydantic_resolve.exceptions import MissingAnnotationError
import pydantic_resolve.utils.conversion as conversion_util
import pydantic_resolve.utils.class_util as class_util
import pydantic_resolve.constant as const
import pydantic_resolve.utils.profile as profile_util


T = TypeVar("T")

class Resolver:
    def __init__(
            self,
            loader_filters: Optional[Dict[Any, Dict[str, Any]]] = None,  # deprecated
            loader_params: Optional[Dict[Any, Dict[str, Any]]] = None,
            global_loader_filter: Optional[Dict[str, Any]] = None,  # deprecated
            global_loader_param: Optional[Dict[str, Any]] = None,
            loader_instances: Optional[Dict[Any, Any]] = None,
            ensure_type=False,
            context: Optional[Dict[str, Any]] = None,
            debug=False,
            enable_from_attribute_in_type_adapter=False,
            ):
        
        self.debug = debug or os.getenv("PYDANTIC_RESOLVE_DEBUG", "false").lower() == "true"

        self.performance = profile_util.Profile() 
        self.loader_instance_cache = {}

        self.ancestor_vars = {}
        self.collector_contextvars = {}
        self.parent_contextvars = {}

        # for dataloader which has class attributes, you can assign the value at here
        if loader_filters:
            warnings.warn('loader_filters is deprecated, use loader_params instead.', DeprecationWarning)
            self.loader_params = loader_filters
        else:
            self.loader_params = loader_params or {}

        # keys in global_loader_filter are mutually exclusive with key-value pairs in loader_filters
        # eg: Resolver(global_loader_filter={'key_a': 1}, loader_filters={'key_a': 1}) will raise exception
        if global_loader_filter:
            warnings.warn('global_loader_filter is deprecated, use global_loader_param instead.', DeprecationWarning)
            self.global_loader_param = global_loader_filter or {}
        else:
            self.global_loader_param = global_loader_param or {}

        # now you can pass your loader instance, Resolver will check `isinstance``
        if loader_instances and self._validate_loader_instance(loader_instances):
            self.loader_instances = loader_instances
        else:
            self.loader_instances = {}

        # only use with pydantic v2
        # for scenario of upgrading from pydantic v1
        # in v1, it supports parsing from another pydantic object which contains all fields the target
        # class required, but in v2, this will raise exception, type adapter by default only support parsing from 
        # dict or pydantic object which is exactly the same with target class
        #
        # class A(BaseModel):
        #   name: str
        #   id: int
        # 
        # class B(BaseModel):
        #   name: str
        # 
        # in pydantc v1, parse_obj_as can parse B from A, but in v2, it will raise exception
        # however, with typeAdapter.validate_python(data, from_attribute=True), it can work
        # the cost is performance (abount 10% overhead), so it is disabled by default
        self.enable_from_attribute_in_type_adapter = enable_from_attribute_in_type_adapter \
            or os.getenv("PYDANTIC_RESOLVE_ENABLE_FROM_ATTRIBUTE", "false").lower() == "true"

        self.ensure_type = ensure_type
        self.context = MappingProxyType(context) if context else None
        self.metadata = {}
        self.object_level_collect_alias_map_store: Dict[int, Dict] = {}

    def _validate_loader_instance(self, loader_instances: Dict[Any, Any]):
        for cls, loader in loader_instances.items():
            if not issubclass(cls, DataLoader):
                raise AttributeError(f'{cls.__name__} must be subclass of DataLoader')
            if not isinstance(loader, cls):
                raise AttributeError(f'{loader.__name__} is not instance of {cls.__name__}')
        return True
    
    def _prepare_collectors(self, node: object, kls: Type):
        alias_map = analysis.generate_alias_map_with_cloned_collector(kls, self.metadata)
        if alias_map:
            # store for later post methods
            self.object_level_collect_alias_map_store[id(node)] = alias_map  

            # expose to descendant
            for alias_name, sign_collector_kv in alias_map.items():
                if not self.collector_contextvars.get(alias_name):
                    self.collector_contextvars[alias_name] = contextvars.ContextVar(alias_name, default={})
                
                current_pair = self.collector_contextvars[alias_name].get()
                if set(sign_collector_kv.keys()) - set(current_pair.keys()):  # update only when new sign is found
                    updated_pair = {**current_pair, **sign_collector_kv}
                    self.collector_contextvars[alias_name].set(updated_pair)

    def _prepare_parent(self, node: object):
        if not self.parent_contextvars.get('parent'):
            self.parent_contextvars['parent'] = contextvars.ContextVar('parent')
        self.parent_contextvars['parent'].set(node)

    def _prepare_expose_fields(self, node: object):
        expose_dict: Optional[dict] = getattr(node, const.EXPOSE_TO_DESCENDANT, None)
        if expose_dict:
            for field, alias in expose_dict.items():  # eg: {'name': 'bar_name'}
                if not self.ancestor_vars.get(alias):
                    self.ancestor_vars[alias] = contextvars.ContextVar(alias)

                try:
                    val = getattr(node, field)
                except AttributeError:
                    raise AttributeError(f'{field} does not existed')

                self.ancestor_vars[alias].set(val)

    def _prepare_ancestor_context(self):
        return {k: v.get() for k, v in self.ancestor_vars.items()}

    def _execute_resolve_method(
            self,
            kls: Type,
            field: str,
            method: Callable):

        params = {}
        resolve_param = analysis.get_resolve_method_param(kls, field, self.metadata)

        if resolve_param['context']:
            params['context'] = self.context
        if resolve_param['ancestor_context']:
            params['ancestor_context'] = self._prepare_ancestor_context()
        if resolve_param['parent']:
            params['parent'] = self.parent_contextvars['parent'].get()
        
        for loader in resolve_param['dataloaders']:
            cache_key = loader['path']
            loader_instance = self.loader_instance_cache[cache_key]
            params[loader['param']] = loader_instance

        return method(**params)
    
    def _execute_post_method(
            self,
            node: object,
            kls: Type,
            kls_path: str,
            field: str,
            method: Callable):

        params = {}
        post_param = analysis.get_post_method_params(kls, field, self.metadata)

        if post_param['context']:
            params['context'] = self.context
        if post_param['ancestor_context']:
            params['ancestor_context'] = self._prepare_ancestor_context()
        if post_param['parent']:
            params['parent'] = self.parent_contextvars['parent'].get()
        
        for loader in post_param['dataloaders']:
            cache_key = loader['path']
            loader_instance = self.loader_instance_cache[cache_key]
            params[loader['param']] = loader_instance

        alias_map = self.object_level_collect_alias_map_store.get(id(node), {})
        if alias_map:
            for collector in post_param['collectors']:
                signature = analysis.get_collector_sign(kls_path, collector)
                alias, param = collector['alias'], collector['param']
                params[param] = alias_map[alias][signature]
        
        return method(**params)

    def _execute_post_default_handler(
            self,
            node: object,
            kls: Type,
            kls_path: str,
            method: Callable):

        params = {}
        post_default_param = analysis.get_post_default_handler_params(kls, self.metadata)

        if post_default_param is None:
            return

        if post_default_param['context']:
            params['context'] = self.context
        if post_default_param['ancestor_context']:
            params['ancestor_context'] = self._prepare_ancestor_context()
        if post_default_param['parent']:
            params['parent'] = self.parent_contextvars['parent'].get()

        alias_map = self.object_level_collect_alias_map_store.get(id(node), {})
        if alias_map:
            for collector in post_default_param['collectors']:
                alias, param = collector['alias'], collector['param']
                signature = (kls_path, const.POST_DEFAULT_HANDLER, param)
                params[param] = alias_map[alias][signature]

        return method(**params)

    def _add_values_into_collectors(self, node: object, kls: Type):
        for field, alias in analysis.get_collector_candidates(kls, self.metadata):
            alias_list = alias if isinstance(alias, (tuple, list)) else (alias,)

            for alias in alias_list:
                for _, instance in self.collector_contextvars[alias].get().items():
                    if isinstance(field, tuple):  # only tuple are allowed to be key
                        val = [getattr(node, f) for f in field]
                    else:
                        val = getattr(node, field)
                    instance.add(val)

    async def _execute_resolve_method_field(
            self, 
            node: object, 
            kls: Type,
            field: str,
            trim_field: str,
            method: Callable):
        if self.ensure_type:
            if not method.__annotations__:
                raise MissingAnnotationError(f'{field}: return annotation is required')

        val = self._execute_resolve_method(kls, field, method)

        while iscoroutine(val) or asyncio.isfuture(val):
            val = await val

        if not getattr(method, const.HAS_MAPPER_FUNCTION, False):  # defined in util.mapper
            val = conversion_util.try_parse_data_to_target_field_type(
                node,
                trim_field,
                val,
                self.enable_from_attribute_in_type_adapter)

        val = await self._traverse(val, node)
        setattr(node, trim_field, val)

    async def _execute_post_method_field(
         self,   
         node: object,
         kls: Type,
         kls_path: str,
         field: str,
         trim_field: str,
         method: Callable
    ):
        val = self._execute_post_method(node, kls, kls_path, field, method)

        while iscoroutine(val) or asyncio.isfuture(val):
            val = await val
            
        if not getattr(method, const.HAS_MAPPER_FUNCTION, False):  # defined in util.mapper
            val = conversion_util.try_parse_data_to_target_field_type(
                node,
                trim_field,
                val,
                self.enable_from_attribute_in_type_adapter)

        setattr(node, trim_field, val)
    
    async def _traverse(self, node: T, parent: object) -> T:
        """
        life cycle:
        - prepare 
            - collectors, 
            - ancestor expose fields 
            - parent
        - execute 
            - resolve method fields/ object fields
            - post method fields
            - post default handler
        - collect
            - values into ancestor collectors
        - return node
        """
        if isinstance(node, (list, tuple)):
            await asyncio.gather(*[self._traverse(t, parent) for t in node])
            return node

        if not analysis.is_acceptable_instance(node):
            return node

        kls = node.__class__
        kls_path = class_util.get_kls_full_name(kls)

        self._prepare_collectors(node, kls)
        self._prepare_expose_fields(node)
        self._prepare_parent(parent)


        if self.debug:
            ancestors = self.ancestor_list.get()
            new_ancestors = ancestors + [node.__class__.__name__]
            self.ancestor_list.set(new_ancestors)
            tid = self.performance.get_timer(new_ancestors).start()


        resolve_tasks = []
        post_tasks = []

        # resolve process
        resolve_fields, object_fields = analysis.get_resolve_fields_and_object_fields_from_object(node, kls, self.metadata)
        for field, resolve_trim_field, method in resolve_fields:
            resolve_tasks.append(self._execute_resolve_method_field(
                node=node, 
                kls=kls, 
                field=field, 
                trim_field=resolve_trim_field, 
                method=method))

        for field, attr_object in object_fields:
            resolve_tasks.append(self._traverse(attr_object, node))

        await asyncio.gather(*resolve_tasks)

        # post process
        for post_field, post_trim_field, method in analysis.get_post_methods(node, kls, self.metadata):
            post_tasks.append(self._execute_post_method_field(
                node=node, 
                kls=kls, 
                kls_path=kls_path, 
                field=post_field, 
                trim_field=post_trim_field,
                method=method))

        await asyncio.gather(*post_tasks)

        default_post_method = getattr(node, const.POST_DEFAULT_HANDLER, None)
        if default_post_method:
            self._execute_post_default_handler(node, kls, kls_path, default_post_method)

        self._add_values_into_collectors(node, kls)

        if self.debug:
            self.performance.get_timer(new_ancestors).end(tid) # type: ignore

        return node

    async def resolve(self, node: T) -> T:
        if isinstance(node, list) and node == []: return node

        root_class = class_util.get_class_of_object(node)
        metadata = analysis.scan_and_store_metadata(root_class)
        self.metadata = analysis.convert_metadata_key_as_kls(metadata)

        self.loader_instance_cache = analysis.validate_and_create_loader_instance(
            self.loader_params,
            self.global_loader_param,
            self.loader_instances,
            self.metadata)
        
        has_context = analysis.has_context(self.metadata)
        if has_context and self.context is None:
            raise AttributeError('context is missing')

        self.ancestor_list = contextvars.ContextVar('ancestor_list', default=[])
            
        await self._traverse(node, None)

        if self.debug:
            self.performance.report()

        return node


import asyncio
import functools
import types
from inspect import iscoroutine
from typing import Any, Callable, Optional, Type, Union
from pydantic import BaseModel, ValidationError, parse_obj_as
from dataclasses import is_dataclass
from pydantic_resolve.compat import PYDANTIC_V2
import pydantic_resolve.constant as const
from pydantic_resolve.utils.class_util import safe_issubclass


if PYDANTIC_V2:
    from pydantic import BaseModel, TypeAdapter, ValidationError
    class TypeAdapterManager:
        apapters = {}
        
        @classmethod
        def get(cls, type):
            adapter = cls.apapters.get(type)
            if adapter:
                return adapter
            else:
                new_adapter = TypeAdapter(type)
                cls.apapters[type] = new_adapter
                return new_adapter

def try_parse_data_to_target_field_type_v1(
        target: object,
        field_name: str,
        data,
        enable_from_attribute=False):
    """
    parse to pydantic or dataclass object
    1. get type of target field
    2. parse
    """
    field_type = None

    # 1. get type of target field
    if isinstance(target, BaseModel):
        _fields = target.__class__.__fields__
        field_type = _fields[field_name].outer_type_

        # handle optional logic
        if data is None and _fields[field_name].required == False:
            return data

    elif is_dataclass(target):
        field_type = target.__class__.__annotations__[field_name]

    # 2. parse
    if field_type:
        try:
            result = parse_obj_as(field_type, data)
            return result
        except ValidationError as e:
            print(f'Warning: type mismatch, pls check the return type for "{field_name}", expected: {field_type}')
            raise e
    else:
        return data  #noqa


def try_parse_data_to_target_field_type_v2(
        target: object,
        field_name: str,
        data,
        enable_from_attribute=False):
    """
    parse to pydantic or dataclass object
    1. get type of target field
    2. parse
    """
    field_type = None

    # from_attribute by default is None
    # if set False it will fail when dealing with namedtuple
    _enable_from_attribute = True if enable_from_attribute else None 

    # 1. get type of target field
    if isinstance(target, BaseModel):
        _fields = target.__class__.model_fields
        field_type = _fields[field_name].annotation

        # handle optional logic
        if data is None and _fields[field_name].is_required() == False:
            return data

    elif is_dataclass(target):
        field_type = target.__class__.__annotations__[field_name]

    # 2. parse
    if field_type:
        try:
            # https://docs.pydantic.dev/latest/concepts/performance/#typeadapter-instantiated-once
            adapter = TypeAdapterManager.get(field_type)
            result = adapter.validate_python(data, from_attributes=_enable_from_attribute)
            return result
        except ValidationError as e:
            print(f'Warning: type mismatch, pls check the return type for "{field_name}", expected: {field_type}')
            raise e

    else:
        return data  #noqa


try_parse_data_to_target_field_type = try_parse_data_to_target_field_type_v2 if PYDANTIC_V2 else try_parse_data_to_target_field_type_v1


def _get_mapping_rule_v1(target, source) -> Optional[Callable]:
    # do noting
    if isinstance(source, target):
        return None

    # pydantic
    if safe_issubclass(target, BaseModel):
        if target.__config__.orm_mode:
            if isinstance(source, dict):
                raise AttributeError(f"{type(source)} -> {target.__name__}: pydantic from_orm can't handle dict object")
            else:
                return lambda t, s: t.from_orm(s)

        if isinstance(source, (dict, BaseModel)):
            return lambda t, s: t.parse_obj(s)

        else:
            raise AttributeError(f"{type(source)} -> {target.__name__}: pydantic can't handle non-dict data")

    # dataclass
    if is_dataclass(target):
        if isinstance(source, dict):
            return lambda t, s: t(**s)

    raise NotImplementedError(f"{type(source)} -> {target.__name__}: faild to get auto mapping rule and execut mapping, use your own rule instead.")


def _get_mapping_rule_v2(target, source) -> Optional[Callable]:
    # do noting
    if isinstance(source, target):
        return None

    # pydantic
    if safe_issubclass(target, BaseModel):
        if target.model_config.get('from_attributes'):
            if isinstance(source, dict):
                raise AttributeError(f"{type(source)} -> {target.__name__}: pydantic from_orm can't handle dict object")
            else:
                return lambda t, s: t.model_validate(s)

        if isinstance(source, dict):
            return lambda t, s: t.model_validate(s)

        if isinstance(source, BaseModel):
            if source.model_config.get('from_attributes'):
                return lambda t, s: t.model_validate(s)
            else:
                return lambda t, s: t(**s.model_dump())

        else:
            raise AttributeError(f"{type(source)} -> {target.__name__}: pydantic can't handle non-dict data")

    # dataclass
    if is_dataclass(target):
        if isinstance(source, dict):
            return lambda t, s: t(**s)

    raise NotImplementedError(f"{type(source)} -> {target.__name__}: faild to get auto mapping rule and execut mapping, use your own rule instead.")


_get_mapping_rule = _get_mapping_rule_v2 if PYDANTIC_V2 else _get_mapping_rule_v1


def _apply_rule(rule: Optional[Callable], target, source: Any, is_list: bool):
    if not rule:  # no change
        return source

    if is_list:
        return [rule(target, s) for s in source]
    else:
        return rule(target, source)


def mapper(func_or_class: Union[Callable, Type]):
    """
    execute post-transform function after the value is reolved
    func_or_class:
        is func: run func
        is class: call auto_mapping to have a try

    @dataclass
    class K:
        id: int

        field: str = ''
        @mapper(lambda x: x.name)
        def resolve_field(self, loader=Loader(field_batch_load_fn)):
            return loader.load(self.id)
    """
    def inner(inner_fn):

        # if mapper provided, auto map from target type will be disabled
        setattr(inner_fn, const.HAS_MAPPER_FUNCTION, True)

        @functools.wraps(inner_fn)
        async def wrap(*args, **kwargs):

            retVal = inner_fn(*args, **kwargs)
            while iscoroutine(retVal) or asyncio.isfuture(retVal):
                retVal = await retVal  # get final result

            if retVal is None:
                return None

            if isinstance(func_or_class, types.FunctionType):
                # manual mapping
                return func_or_class(retVal)
            else:
                # auto mapping
                if isinstance(retVal, list):
                    if retVal:
                        rule = _get_mapping_rule(func_or_class, retVal[0])
                        return _apply_rule(rule, func_or_class, retVal, True)
                    else:
                        return retVal  # return []
                else:
                    rule = _get_mapping_rule(func_or_class, retVal)
                    return _apply_rule(rule, func_or_class, retVal, False)
        return wrap
    return inner


from typing import Callable, Type


def replace_method(cls: Type, cls_name: str, func_name: str, func: Callable):
    """test-only"""
    KLS = type(cls_name, (cls,), {func_name: func})
    return KLS


from collections import defaultdict
from typing import Callable, DefaultDict, Iterator, List, Mapping, Optional, Sequence, TypeVar
from aiodataloader import DataLoader

T = TypeVar("T")
V = TypeVar("V")

def build_list(items: Sequence[T], keys: List[V], get_pk: Callable[[T], V]) -> Iterator[List[T]]:
    """
    helper function to build return list data required by aiodataloader
    """
    dct: DefaultDict[V, List[T]] = defaultdict(list)
    for item in items:
        _key = get_pk(item)
        dct[_key].append(item)
    results = (dct.get(k, []) for k in keys)
    return results


def build_object(items: Sequence[T], keys: List[V], get_pk: Callable[[T], V]) -> Iterator[Optional[T]]:
    """
    helper function to build return object data required by aiodataloader
    """
    dct: Mapping[V, T] = {}
    for item in items:
        _key = get_pk(item)
        dct[_key] = item
    results = (dct.get(k, None) for k in keys)
    return results


def copy_dataloader_kls(name, loader_kls):
    """
    quickly copy from an existed DataLoader class
    usage:
    SeniorMemberLoader = copy_dataloader('SeniorMemberLoader', ul.UserByLevelLoader)
    JuniorMemberLoader = copy_dataloader('JuniorMemberLoader', ul.UserByLevelLoader)
    """
    return type(name, loader_kls.__bases__, dict(loader_kls.__dict__))


class StrictEmptyLoader(DataLoader):
    async def batch_load_fn(self, keys):
        """it should not be triggered, otherwise will raise Exception"""
        raise ValueError('EmptyLoader should load from pre loaded data')


class ListEmptyLoader(DataLoader):
    async def batch_load_fn(self, keys):
        dct = {}
        return [dct.get(k, []) for k in keys]


class SingleEmptyLoader(DataLoader):
    async def batch_load_fn(self, keys):
        dct = {}
        return [dct.get(k, None) for k in keys]


def generate_strict_empty_loader(name):
    """generated Loader will raise ValueError if not found"""
    return type(name, StrictEmptyLoader.__bases__, dict(StrictEmptyLoader.__dict__))  #noqa


def generate_list_empty_loader(name):
    """generated Loader will return [] if not found"""
    return type(name, ListEmptyLoader.__bases__, dict(ListEmptyLoader.__dict__))  #noqa


def generate_single_empty_loader(name):
    """generated Loader will return None if not found"""
    return type(name, SingleEmptyLoader.__bases__, dict(SingleEmptyLoader.__dict__))  #noqa


import functools
from dataclasses import is_dataclass, fields, MISSING
from typing import Type, get_type_hints, List, Tuple
import pydantic_resolve.constant as const
from pydantic_resolve.compat import PYDANTIC_V2
import pydantic_resolve.utils.class_util as class_util
from pydantic_resolve.analysis import is_acceptable_kls
from pydantic_resolve.utils.types import get_type, _is_optional, get_core_types
from pydantic import BaseModel


# ----------------------- rebuild -----------------------
def rebuild_v1(kls):
    kls.update_forward_refs()


def rebuild_v2(kls):
    kls.model_rebuild()


rebuild = rebuild_v2 if PYDANTIC_V2 else rebuild_v1


# ---------------------- ensure_subset ------------------
def ensure_subset_v1(base):
    """
    used with pydantic class or dataclass to make sure a class's field is 
    subset of target class
    """
    def wrap(kls):
        is_base_pydantic = safe_issubclass(base, BaseModel)
        is_kls_pydantic = safe_issubclass(kls, BaseModel)
        is_base_dataclass = is_dataclass(base)
        is_kls_dataclass = is_dataclass(kls)

        if is_base_pydantic and is_kls_pydantic:
            @functools.wraps(kls)
            def inner():
                for k, field in kls.__fields__.items():
                    if field.required:
                        base_field = base.__fields__.get(k)
                        if not base_field:
                            raise AttributeError(f'{k} not existed in {base.__name__}.')
                        if base_field and base_field.type_ != field.type_:
                            raise AttributeError(f'type of {k} not consistent with {base.__name__}')
                return kls
            return inner()
        elif is_base_dataclass and is_kls_dataclass:
            @functools.wraps(kls)
            def inner():
                base_fields = {f.name: f.type for f in fields(base)}
                for f in fields(kls):
                    has_default = is_dataclass_field_has_default_value(f)
                    if not has_default:
                        if f.name not in base_fields:
                            raise AttributeError(f'{f.name} not existed in {base.__name__}.')
                        if base_fields[f.name] != f.type:
                            raise AttributeError(f'type of {f.name} not consistent with {base.__name__}')
                return kls
            return inner()
        else:
            raise TypeError('base and kls should both be Pydantic BaseModel or both be dataclass')
    return wrap


def ensure_subset_v2(base):
    """
    used with pydantic class or dataclass to make sure a class's field is 
    subset of target class

    for pydantic v2, subclass with Optional[T] but without default value will raise exception

    @class_util.ensure_subset(Base)
    class ChildB1(BaseModel):
        a: str
        d: Optional[int]
    
    this will raise

    @class_util.ensure_subset(Base)
    class ChildB1(BaseModel):
        a: str
        d: Optional[int] = 0 
    
    this is ok
    """
    def wrap(kls):
        is_base_pydantic = safe_issubclass(base, BaseModel)
        is_kls_pydantic = safe_issubclass(kls, BaseModel)
        is_base_dataclass = is_dataclass(base)
        is_kls_dataclass = is_dataclass(kls)

        if is_base_pydantic and is_kls_pydantic:
            @functools.wraps(kls)
            def inner():
                for k, field in kls.model_fields.items():
                    if field.is_required():
                        base_field = base.model_fields.get(k)
                        if not base_field:
                            raise AttributeError(f'{k} not existed in {base.__name__}.')
                        if base_field and base_field.annotation != field.annotation:
                            raise AttributeError(f'type of {k} not consistent with {base.__name__}')
                return kls
            return inner()

        elif is_base_dataclass and is_kls_dataclass:
            @functools.wraps(kls)
            def inner():
                base_fields = {f.name: f.type for f in fields(base)}
                for f in fields(kls):
                    has_default = is_dataclass_field_has_default_value(f)
                    if not has_default:
                        if f.name not in base_fields:
                            raise AttributeError(f'{f.name} not existed in {base.__name__}.')
                        if base_fields[f.name] != f.type:
                            raise AttributeError(f'type of {f.name} not consistent with {base.__name__}')
                return kls
            return inner()
        else:
            raise TypeError('base and kls should both be Pydantic BaseModel or both be dataclass')
    return wrap


ensure_subset = ensure_subset_v2 if PYDANTIC_V2 else ensure_subset_v1


def _get_pydantic_field_items_v1(kls):
    return kls.__fields__.items()


def _get_pydantic_field_items_v2(kls):
    return kls.model_fields.items()


get_pydantic_field_items = _get_pydantic_field_items_v2 if PYDANTIC_V2 else _get_pydantic_field_items_v1


def _get_pydantic_field_keys_v1(kls) -> str:
    return kls.__fields__.keys()


def _get_pydantic_field_keys_v2(kls) -> str:
    return kls.model_fields.keys()


get_pydantic_field_keys = _get_pydantic_field_keys_v2 if PYDANTIC_V2 else _get_pydantic_field_keys_v1


def _get_pydantic_field_values_v1(kls):
    return kls.__fields__.values()


def _get_pydantic_field_values_v2(kls):
    return kls.model_fields.values()


get_pydantic_field_values = _get_pydantic_field_values_v2 if PYDANTIC_V2 else _get_pydantic_field_values_v1


def _is_pydantic_field_required_v1(field):
    return field.required

def _is_pydantic_field_required_v2(field):
    return field.is_required()

is_pydantic_field_required_field = _is_pydantic_field_required_v2 if PYDANTIC_V2 else _is_pydantic_field_required_v1


def get_pydantic_fields(kls):
    items = class_util.get_pydantic_field_items(kls)

    for name, v in items:
        t = get_type(v)
        shelled_types = get_core_types(t)
        allowed_types = [st for st in shelled_types if is_acceptable_kls(st)]
        if allowed_types:
            yield (name, allowed_types)  # type_ is the most inner type


def get_dataclass_fields(kls):
    for name, v in kls.__annotations__.items():
        shelled_types = get_core_types(v)
        allowed_types = [st for st in shelled_types if is_acceptable_kls(st)]
        if allowed_types:
            yield (name, allowed_types)


def get_class_of_object(target):
    if isinstance(target, list):
        return target[0].__class__
    else:
        return target.__class__


def is_dataclass_field_has_default_value(field):
    if field.default is not MISSING or field.default_factory is not MISSING:
        return True

    typ = field.type
    return _is_optional(typ)


def update_forward_refs(kls):
    def update_pydantic_forward_refs(kls: Type[BaseModel]):
        """
        recursively update refs.
        """
        if getattr(kls, const.PYDANTIC_FORWARD_REF_UPDATED, False):
            return

        rebuild(kls)

        setattr(kls, const.PYDANTIC_FORWARD_REF_UPDATED, True)
        
        values = get_pydantic_field_values(kls)

        for field in values:
            shelled_types = get_core_types(get_type(field))
            for shelled_type in shelled_types:
                update_forward_refs(shelled_type)

    def update_dataclass_forward_refs(kls):
        if not getattr(kls, const.DATACLASS_FORWARD_REF_UPDATED, False):
            anno = get_type_hints(kls)
            kls.__annotations__ = anno
            setattr(kls, const.DATACLASS_FORWARD_REF_UPDATED, True)

            for _, v in kls.__annotations__.items():
                shelled_types = get_core_types(v)
                for shelled_type in shelled_types:
                    update_forward_refs(shelled_type)

    if safe_issubclass(kls, BaseModel):
        update_pydantic_forward_refs(kls)

    if is_dataclass(kls):
        update_dataclass_forward_refs(kls)


def get_kls_full_name(kls):
    return f'{kls.__module__}.{kls.__qualname__}'


def get_fields_default_value_not_provided(cls: Type) -> List[Tuple[str, bool]]:  # field name, has default value
    """
    return class field which do not have a default value.

    class MyClass:
        a: int
        b: int = 1

    print(hasattr(MyClass, 'a'))  # False
    print(hasattr(MyClass, 'b'))  # True
    """
    anno = cls.__dict__.get('__annotations__') or {}
    return [(k, hasattr(cls, k)) for k in anno.keys()]


def safe_issubclass(kls, classinfo):
    try:
        return issubclass(kls, classinfo)
    except TypeError:
        return False


import logging

def get_logger(name: str):
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)
    handler = logging.StreamHandler()
    formatter = logging.Formatter("[%(asctime)s] [%(name)s] [%(levelname)s] %(message)s")
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return logger



RESOLVE_PREFIX = 'resolve_'
POST_PREFIX = 'post_'
PYDANTIC_FORWARD_REF_UPDATED = '__pydantic_resolve_forward_refs_updated__'
DATACLASS_FORWARD_REF_UPDATED = '__pydantic_resolve_dataclass_forward_refs_updated__'
HAS_MAPPER_FUNCTION = '__pydantic_resolve_mapper_provided__'
POST_DEFAULT_HANDLER = 'post_default_handler'
EXPOSE_TO_DESCENDANT = '__pydantic_resolve_expose__'
COLLECTOR_CONFIGURATION = '__pydantic_resolve_collect__'



import time
import math
from typing import List
from uuid import uuid1
from pydantic_resolve.utils.logger import get_logger

profile_logger = get_logger(__name__)

class Profile():
    def __init__(self):
        self.full_path_timer = {}
    
    def get_timer(self, path: List[str]):
        key = '.'.join(path)
        if key not in self.full_path_timer:
            self.full_path_timer[key] = Timer(key)
        return self.full_path_timer[key]
    
    def __repr__(self) -> str:
        def format_key(key, longest):
            return key + ' ' * (longest - len(key))

        longest = max([len(key) for key in self.full_path_timer.keys()])
        items = self.full_path_timer.items()
        items = sorted(items, key=lambda x: x[0])
        items = [f'{format_key(item[0], longest)}: {item[1]}' for item in items]
        return '\n'.join(items)
    
    def report(self):
        profile_logger.debug('\n' + self.__repr__())
    

class Timer():
    def __init__(self, name: str):
        self.name = name
        self._max = 0
        self._min = math.inf
        self.timeset = {} 
        self.records = []
    
    def start(self):
        id = uuid1()
        self.timeset[id] = time.time()
        return id
    
    def end(self, id: str):
        start = self.timeset[id]
        t = self.to_ms(time.time() - start)

        self.records.append(t)
        self._max = max(self._max, t)
        self._min = min(self._min, t)
    
    @property
    def average(self):
        try: 
            return sum(self.records) / len(self.records)
        except ZeroDivisionError:
            return 0
    
    @property
    def max(self):
        return self._max
    
    @property
    def min(self):
        return self._min
    
    def to_ms(self, t):
        return t * 1000
    
    def __repr__(self) -> str:
        return f'avg: {self.average:.1f}ms, max: {self.max:.1f}ms, min: {self.min:.1f}ms'


from typing import Type, Union, List
from pydantic_resolve.compat import PYDANTIC_V2, OVER_PYTHON_3_7

if OVER_PYTHON_3_7:
    from typing import get_origin, get_args


def _is_optional_3_7(annotation):
    annotation_origin = getattr(annotation, "__origin__", None)
    return (
        annotation_origin == Union
        and len(getattr(annotation, "__args__", ())) == 2
        and (
            annotation.__args__[1] is type(None)  # noqa: E721
            or annotation.__args__[0] is type(None)  # noqa: E721
        )
    )


def _is_optional_3_x(annotation):
    origin = get_origin(annotation)
    args = get_args(annotation)
    if origin is Union and type(None) in args:
        return True
    return False


_is_optional = _is_optional_3_x if OVER_PYTHON_3_7 else _is_optional_3_7
_get_origin = get_origin if OVER_PYTHON_3_7 else lambda x: getattr(x, "__origin__", None)
_get_args = get_args if OVER_PYTHON_3_7 else lambda x: getattr(x, "__args__", ())


def _is_list(annotation):
    return _get_origin(annotation) in (list, List)


def shelling_type(tp):
    while _is_optional(tp) or _is_list(tp):
        tp = tp.__args__[0]
    return tp


def get_core_types(tp):
    """
    - get the core type
    - always return a tuple of core types
    """
    if tp is type(None):
        return tuple()

    # 1. Unwrap list layers
    def _shell_list(_tp):
        while _is_list(_tp):
            args = getattr(_tp, "__args__", ())
            if args:
                _tp = args[0]
            else:
                break
        return _tp
    
    tp = _shell_list(tp)

    if tp is type(None): # check again
        return tuple()

    while True:
        orig = _get_origin(tp)
        if orig is Union:
            args = list(_get_args(tp))
            non_none = [a for a in args if a is not type(None)]  # noqa: E721
            has_none = len(non_none) != len(args)
            # Optional[T] case -> keep unwrapping
            if has_none and len(non_none) == 1:
                tp = non_none[0]
                tp = _shell_list(tp)
                continue
            # Union (with or without None) -> return tuple of real types
            if non_none:
                return tuple(non_none)
            return tuple()
        break

    # single concrete type
    return (tp,)


def get_class_field_annotations(cls: Type):
    anno = cls.__dict__.get('__annotations__') or {}
    return anno.keys()


def _get_type_v1(v):
    return v.type_


def _get_type_v2(v):
    return v.annotation


get_type = _get_type_v2 if PYDANTIC_V2 else _get_type_v1



from typing import Any, Callable, Optional
from aiodataloader import DataLoader

class Depends:
    def __init__(
        self,
        dependency: Optional[Callable[..., Any]] = None,
    ):
        self.dependency = dependency


def LoaderDepend(  # noqa: N802
    dependency: Optional[Callable[..., Any]] = None,
) -> DataLoader:
    return Depends(dependency=dependency)


Loader = LoaderDepend


import copy
import inspect
from typing import List, Type, Dict, Optional, Tuple
from inspect import isfunction, isclass
from collections import defaultdict
from aiodataloader import DataLoader
from pydantic import BaseModel
from dataclasses import is_dataclass, fields as dc_fields

import pydantic_resolve.constant as const
import pydantic_resolve.utils.class_util as class_util
from pydantic_resolve.utils.collector import ICollector
from pydantic_resolve.utils.depend import Depends
import pydantic_resolve.utils.params as params_util
from pydantic_resolve.exceptions import ResolverTargetAttrNotFound, LoaderFieldNotProvidedError, MissingCollector
import sys

if sys.version_info >= (3, 8):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict

class DataLoaderType(TypedDict):
    param: str
    kls: Type
    path: str
    request_type: List[Type]

class CollectorType(TypedDict):
    field: str 
    param: str
    instance: object
    alias: str
    
class ResolveMethodType(TypedDict):
    trim_field: str
    context: bool
    parent: bool
    ancestor_context: bool
    dataloaders: List[DataLoaderType]

class PostMethodType(TypedDict):
    trim_field: str
    context: bool
    parent: bool
    ancestor_context: bool
    dataloaders: List[DataLoaderType]
    collectors: List[CollectorType] 

class PostDefaultHandlerType(TypedDict):
    context: bool
    parent: bool
    ancestor_context: bool
    collectors: List[CollectorType]

class KlsMetaType(TypedDict):
    resolve: List[str]
    post: List[str]
    resolve_params: Dict[str, ResolveMethodType]
    post_params: Dict[str, PostMethodType]
    post_default_handler_params: Optional[PostDefaultHandlerType]
    raw_object_fields: List[str] # store the original field names
    object_fields: List[str]
    expose_dict: dict
    collect_dict: dict
    kls: Type
    has_context: bool
    should_traverse: bool

class LoaderQueryMetaRequestType(TypedDict):
    name: Type
    fields: List[str]

class LoaderQueryMeta(TypedDict):
    required_types: List
    fields: List[str]

MetaType = Dict[str, KlsMetaType]

class MappedMetaMemberType(KlsMetaType):
    kls_path: str
    alias_map_proto: Dict[str, Dict[Tuple[str, str, str], object]]

MappedMetaType = Dict[Type, MappedMetaMemberType]


def _scan_resolve_method(method, field: str, request_types: List[Type]) -> ResolveMethodType:
    result: ResolveMethodType = {
        'trim_field': field.replace(const.RESOLVE_PREFIX, ''),
        'context': False,
        'parent': False,
        'ancestor_context': False,
        'dataloaders': []  # collect func or class, do not create instance
    }
    signature = inspect.signature(method)

    if signature.parameters.get('context'):
        result['context'] = True

    if signature.parameters.get('ancestor_context'):
        result['ancestor_context'] = True

    if signature.parameters.get('parent'):
        result['parent'] = True

    for name, param in signature.parameters.items():
        if isinstance(param.default, Depends):
            info: DataLoaderType = { 
                'param': name,
                'kls': param.default.dependency,  # for later initialization
                'path': class_util.get_kls_full_name(param.default.dependency),
                'request_type': request_types
            }
            result['dataloaders'].append(info)

    for name, param in signature.parameters.items():
        if isinstance(param.default, ICollector):
            raise AttributeError("Collector is not available in resolve_method")

    return result


def _scan_post_method(method, field: str, request_types: List[Type]) -> PostMethodType:
    result: PostMethodType = {
        'trim_field': field.replace(const.POST_PREFIX, ''),
        'context': False,
        'ancestor_context': False,
        'parent': False,
        'dataloaders': [],
        'collectors': []
    }
    signature = inspect.signature(method)

    if signature.parameters.get('context'):
        result['context'] = True

    if signature.parameters.get('ancestor_context'):
        result['ancestor_context'] = True

    if signature.parameters.get('parent'):
        result['parent'] = True

    
    for name, param in signature.parameters.items():
        if isinstance(param.default, Depends):
            loader_info: DataLoaderType = { 
                'param': name,
                'kls': param.default.dependency,  # for later initialization
                'path': class_util.get_kls_full_name(param.default.dependency),
                'request_type': request_types
            }
            result['dataloaders'].append(loader_info)

        if isinstance(param.default, ICollector):
            info: CollectorType = {
                'field': field,
                'param': name,
                'instance': param.default,
                'alias': param.default.alias 
            }
            result['collectors'].append(info)

            
    return result


def _scan_post_default_handler(method) -> PostDefaultHandlerType:
    result: PostDefaultHandlerType = {
        'context': False,
        'ancestor_context': False,
        'parent': False,
        'collectors': []
    }
    signature = inspect.signature(method)

    if signature.parameters.get('context'):
        result['context'] = True

    if signature.parameters.get('ancestor_context'):
        result['ancestor_context'] = True

    if signature.parameters.get('parent'):
        result['parent'] = True

    for name, param in signature.parameters.items():
        if isinstance(param.default, ICollector):
            info: CollectorType = {
                'field': const.POST_DEFAULT_HANDLER,
                'param': name,
                'instance': param.default,
                'alias': param.default.alias 
            }
            result['collectors'].append(info)

    return result


def validate_and_create_loader_instance(
        loader_params,
        global_loader_param,
        loader_instances,
        metadata: MappedMetaType):
    """
    return loader_instance_cache

    validate: whether loader params are missing
    create: 
        - func
        - loader class
            - no param
            - has param
    """
    # fetch all loaders
    def _get_all_loaders_from_meta(metadata):
        for _, kls_info in metadata.items():
            for _, resolve_info in kls_info['resolve_params'].items():
                for loader in resolve_info['dataloaders']:
                    yield loader

            for _, post_info in kls_info['post_params'].items():
                for loader in post_info['dataloaders']:
                    yield loader
    
    def _create_instance(loader):
        """
        1. is class?
            - validate params
        2. is func """
        loader_kls, path = loader['kls'], loader['path']
        if isclass(loader_kls):
            loader = loader_kls()
            param_config = params_util.merge_dicts(
                global_loader_param,
                loader_params.get(loader_kls, {}))

            for field, has_default in class_util.get_fields_default_value_not_provided(loader_kls):
                try:
                    if has_default and field not in param_config:
                        continue

                    value = param_config[field]
                    setattr(loader, field, value)
                except KeyError:
                    raise LoaderFieldNotProvidedError(f'{path}.{field} not found in Resolver()')
            return loader
        else:
            loader = DataLoader(batch_load_fn=loader_kls)  # type:ignore
            return loader
    
    def _get_all_fields(kls):
        if class_util.safe_issubclass(kls, BaseModel):
            return list(class_util.get_pydantic_field_keys(kls))

        elif is_dataclass(kls):
            return [f.name for f in dc_fields(kls)]

        else:
            raise AttributeError('invalid type: should be pydantic object or dataclass object')  #noqa

    def _generate_meta(types: List[List[Type]]):
        _fields = set()
        meta: LoaderQueryMeta = {
            'fields': [],
            'request_types': []
        }

        for tt in types:
            for t in tt:
                fields = _get_all_fields(t)
                meta['request_types'].append(dict(name=t, fields=fields))
                _fields.update(fields)
        meta['fields'] = list(_fields)
        return meta
    
    cache = {}
    request_info = {}
    
    # create instance
    for loader in _get_all_loaders_from_meta(metadata):
        loader_kls, path = loader['kls'], loader['path']

        if path in cache: continue

        if loader_instances.get(loader_kls):  # if instance already exists
            cache[path] = loader_instances.get(loader_kls)
            continue

        cache[path] = _create_instance(loader)    

    # prepare query meta
    for loader in _get_all_loaders_from_meta(metadata):
        kls, path = loader['request_type'], loader['path']

        if kls is None:
            continue

        if path in request_info and kls not in request_info[path]:
            request_info[path].append(kls)
        else: 
            request_info[path] = [kls]

    # combine together
    for path, instance in cache.items():
        if request_info.get(path) is None:
            continue

        instance._query_meta = _generate_meta(request_info[path])

    return cache


def scan_and_store_metadata(root_class: Type) -> MetaType:

    class_util.update_forward_refs(root_class)
    expose_set = set()  # for validation
    collect_set = set()  # for validation
    metadata: MetaType = {}

    def _get_request_type_for_loader(object_field_pairs, field_name: str):
        return object_field_pairs.get(field_name)

    def _get_all_fields_and_object_fields(kls):
        if class_util.safe_issubclass(kls, BaseModel):
            all_fields = set(class_util.get_pydantic_field_keys(kls))
            object_fields = list(class_util.get_pydantic_fields(kls))  # dive and recursively analysis
        elif is_dataclass(kls):
            all_fields = set([f.name for f in dc_fields(kls)])
            object_fields = list(class_util.get_dataclass_fields(kls))
        else:
            raise AttributeError('invalid type: should be pydantic object or dataclass object')  #noqa
        return all_fields, object_fields, { x:y for x, y in object_fields}
    
    def _has_post_default_handler(kls):
        fields = dir(kls)
        fields = [f for f in fields if f == const.POST_DEFAULT_HANDLER and isfunction(getattr(kls, f))]
        return len(fields) > 0

    def _get_resolve_and_post_fields(kls):
        fields = dir(kls)

        resolve_fields = [f for f in fields if f.startswith(const.RESOLVE_PREFIX) and isfunction(getattr(kls, f))]
        post_fields = [f for f in fields if f.startswith(const.POST_PREFIX) 
                                                and f != const.POST_DEFAULT_HANDLER
                                                and isfunction(getattr(kls, f))]
        fields_with_resolver = { field.replace(const.RESOLVE_PREFIX, '') for field in resolve_fields }
        return resolve_fields, post_fields, fields_with_resolver
    
    def _validate_resolve_and_post_fields(resolve_fields, post_fields, all_fields):
        for field in resolve_fields:
            resolve_field = field.replace(const.RESOLVE_PREFIX, '')
            if resolve_field not in all_fields:
                raise ResolverTargetAttrNotFound(f"attribute {resolve_field} not found")

        for field in post_fields:
            post_field = field.replace(const.POST_PREFIX, '')
            if post_field not in all_fields:
                raise ResolverTargetAttrNotFound(f"attribute {post_field} not found")

    def _validate_expose(expose_dict, kls_name):
        if type(expose_dict) is not dict:
            raise AttributeError(f'{const.EXPOSE_TO_DESCENDANT} is not dict')
        for _, v in expose_dict.items():
            if v in expose_set:
                raise AttributeError(f'Expose alias name conflicts, please check: {kls_name}')
            expose_set.add(v)
    
    def _add_collector_info(post_params):
        for _, param in post_params.items():
            for collector in param['collectors']:
                collect_set.add(collector['alias'])

    def _add_collector_info_for_default_handler(post_default_params):
        if post_default_params is not None:
            for collector in post_default_params['collectors']:
                collect_set.add(collector['alias'])

    def _validate_collector(collect_dict, kls_name):
        """
        validate if the collector is declared in ancestor nodes.

        available format of `__pydantic_resolve_collect__`

        class A(BaseModel):
            __pydantic_resolve_collect__ = {
                'name': 'collector_a',                                  # send name to collector_a
                ('name', 'age'): 'collector_b',                         # send ('name', 'age') to collector_b
                'name': ('collector_c', 'collector_d'),                 # send name to both collector_c and d
                ('name', 'age'): ('collector_e', 'collector_f'),        # send ('name', 'age') to both collector_e and f
            }
        """
        if type(collect_dict) is not dict:
            raise AttributeError(f'{const.COLLECTOR_CONFIGURATION} is not dict')

        for _, collector in collect_dict.items():
            collectors = collector if isinstance(collector, (list, tuple)) else (collector,)
            for c in collectors:
                if c not in collect_set:
                    raise MissingCollector(f'Collector alias name not found in ancestor, please check: {kls_name}')
    
    def _has_config(info: KlsMetaType):
        result = len(info['resolve']) > 0 or \
                len(info['post']) > 0 or \
                len(info['collect_dict']) > 0 or \
                len(info['expose_dict']) > 0 or \
                info['post_default_handler_params'] is not None
        return result

    def _populate_ancestors(parents):
        for field, kls_name in parents:
            # normally, array size is small
            if field in metadata[kls_name]['raw_object_fields'] and \
                field not in metadata[kls_name]['object_fields']:
                metadata[kls_name]['object_fields'].append(field)
            metadata[kls_name]['should_traverse'] = True

    def walker(kls, ancestors: List[Tuple[str, str]]):
        kls_name = class_util.get_kls_full_name(kls)
        hit = metadata.get(kls_name)
        if hit:
            # if populated by previous node, or self has_config
            if hit['should_traverse'] or _has_config(hit):
                _populate_ancestors(ancestors)
            return

        # - prepare fields, with resolve_, post_ reserved
        all_fields, object_fields, object_field_pairs = _get_all_fields_and_object_fields(kls)
        resolve_fields, post_fields, fields_with_resolver = _get_resolve_and_post_fields(kls)
        _validate_resolve_and_post_fields(resolve_fields, post_fields, all_fields)
        object_fields_without_resolver = [a[0] for a in object_fields if a[0] not in fields_with_resolver] 

        # - scan expose and collect (__pydantic_resolve_xxx__)
        expose_dict = getattr(kls, const.EXPOSE_TO_DESCENDANT, {})
        collect_dict = getattr(kls, const.COLLECTOR_CONFIGURATION, {})
        _validate_expose(expose_dict, kls_name)

        resolve_params = {
            field: _scan_resolve_method(
                    getattr(kls, field),
                    field,
                    _get_request_type_for_loader(
                        object_field_pairs, 
                        field.replace(const.RESOLVE_PREFIX, ''))) 
                for field in resolve_fields
        }
        post_params = {
            field: _scan_post_method(
                    getattr(kls, field),
                    field,
                    _get_request_type_for_loader(
                        object_field_pairs, 
                        field.replace(const.POST_PREFIX, ''))) 
                for field in post_fields
        }
        post_default_handler_params = _scan_post_default_handler(getattr(kls, const.POST_DEFAULT_HANDLER)) if _has_post_default_handler(kls) else None

        # check context
        resolve_context = any([p['context'] for p in resolve_params.values()])
        post_context = any([p['context'] for p in post_params.values()])
        post_default_context = post_default_handler_params['context'] if post_default_handler_params else False
        has_context = resolve_context or post_context or post_default_context

        # check collector
        _add_collector_info(post_params)
        _add_collector_info_for_default_handler(post_default_handler_params)
        _validate_collector(collect_dict, kls_name)

        info: KlsMetaType = {
            'resolve': resolve_fields,
            'resolve_params': resolve_params,
            'post': post_fields,
            'post_params': post_params,
            'post_default_handler_params': post_default_handler_params,
            'raw_object_fields': object_fields_without_resolver,
            'object_fields': [],
            'expose_dict': expose_dict,
            'collect_dict': collect_dict,
            'kls': kls,
            'has_context': has_context,
            'should_traverse': False  
        }
        metadata[kls_name] = info

        # object fields
        for field, shelled_types in (obj for obj in object_fields if obj[0] in object_fields_without_resolver):
            for shelled_type in shelled_types:
                walker(shelled_type, ancestors + [(field, kls_name)])

        # resolve fields
        for field, shelled_types in (obj for obj in object_fields if obj[0] in fields_with_resolver):
            for shelled_type in shelled_types:
                walker(shelled_type, ancestors + [(field, kls_name)])

        should_traverse = info['should_traverse'] or _has_config(info)

        if should_traverse:
            _populate_ancestors(ancestors)

        info['should_traverse'] = should_traverse

    walker(root_class, [])
    return metadata


def convert_metadata_key_as_kls(metadata: MetaType) -> MappedMetaType:
    """ use kls as map key for performance """
    kls_metadata = {}
    for k, v in metadata.items():
        _kls_meta: MappedMetaMemberType = {
            **v, 
            'kls_path':k,
            'alias_map_proto': {}
        }
        alias_map = _calc_alias_map_from_collectors(_kls_meta)
        _kls_meta['alias_map_proto'] = alias_map
        kls_metadata[v['kls']] = _kls_meta
    
    return kls_metadata


def _calc_alias_map_from_collectors(kls_meta: MappedMetaMemberType):
    post_params = kls_meta['post_params']
    kls_path = kls_meta['kls_path']

    alias_map = defaultdict(dict)

    def add_collector(collector: CollectorType):
        """
        class A(BaseModel):
            def post_name(collector=Collector('name')):
                ...
        kls_path: module_name.A
        field: post_name
        param: collector_a

        sign = ('module_name.A', 'post_name', 'collector_a')

        return { 'name': { sign: collector_instance } }
        """
        sign = get_collector_sign(kls_path, collector)
        alias_map[collector['alias']][sign] = collector['instance']

    for _, param in post_params.items():
        for collector in param['collectors']:
            add_collector(collector)

    # post_default_handler
    post_default_handler_params = kls_meta['post_default_handler_params']
    if post_default_handler_params:
        for collector in post_default_handler_params['collectors']:
            add_collector(collector)

    return alias_map


def is_acceptable_kls(kls: Type) -> bool:
    return class_util.safe_issubclass(kls, BaseModel) or is_dataclass(kls)


def is_acceptable_instance(target: object):
    """ check whether target is Pydantic object or Dataclass object """
    return isinstance(target, BaseModel) or is_dataclass(target)


def get_resolve_fields_and_object_fields_from_object(node: object, kls: Type, mapped_metadata: MappedMetaType):
    """
    resolve_fields: a field with resolve_field method
    object_fields: a field without resolve_field method but is an object (should traversal)
    """
    resolve_fields, object_fields = [], []
    kls_meta = mapped_metadata.get(kls)

    if kls_meta is None:
        return [], []

    for resolve_field in kls_meta['resolve']:
        attr = getattr(node, resolve_field)
        trim_field = kls_meta['resolve_params'][resolve_field]['trim_field']
        resolve_fields.append((resolve_field, trim_field, attr))

    for attr_name in kls_meta['object_fields']:
        attr = getattr(node, attr_name)
        object_fields.append((attr_name, attr))

    return resolve_fields, object_fields


def get_post_methods(node: object, kls: Type, mapped_metadata: MappedMetaType):
    kls_meta = mapped_metadata.get(kls)

    if kls_meta is None:
        return []

    for post_field in kls_meta['post']:
        attr = getattr(node, post_field)
        trim_field = kls_meta['post_params'][post_field]['trim_field']
        yield post_field, trim_field, attr


def get_resolve_method_param(kls: Type, resolve_field: str, mapped_metadata: MappedMetaType):
    kls_meta = mapped_metadata.get(kls, {})
    return kls_meta['resolve_params'][resolve_field]


def get_post_method_params(kls, post_field, mapped_metadata: MappedMetaType):
    kls_meta = mapped_metadata.get(kls, {})
    return kls_meta['post_params'][post_field]


def get_post_default_handler_params(kls, mapped_metadata: MappedMetaType):
    kls_meta = mapped_metadata.get(kls, {})
    return kls_meta['post_default_handler_params']


def get_collector_sign(kls_path: str, collector: CollectorType):
    return (kls_path, collector['field'], collector['param'])


def generate_alias_map_with_cloned_collector(kls: Type, mapped_metadata: MappedMetaType):
    kls_meta = mapped_metadata.get(kls, {})
    return { alias: {
        sign: copy.deepcopy(collector) for sign, collector in v.items()
    } for alias, v in kls_meta['alias_map_proto'].items()}


def get_collector_candidates(kls, metadata: MappedMetaType):
    kls_meta = metadata.get(kls, {})
    collect_dict = kls_meta['collect_dict']

    for field, alias in collect_dict.items():
        yield field, alias


def has_context(mapped_metadata: MappedMetaType):
    return any([ m['has_context'] for m in mapped_metadata.values()])


import abc
from typing import Any, List, Union

class ICollector(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def __init__(self, alias: str):
        self.alias = alias

    @abc.abstractmethod
    def add(self, val):
        """how to add new element(s)"""

    @abc.abstractmethod
    def values(self) -> Any:
        """get result"""


class Collector(ICollector):
    def __init__(self, alias: str, flat: bool=False):
        super().__init__(alias)
        self.flat = flat
        self.val = []

    def add(self, val: Union[Any, List[Any]]) -> None:
        if self.flat:
            if isinstance(val, list):
                self.val.extend(val)
            else:
                raise TypeError('if flat, target should be list')
        else:
            self.val.append(val)

    def values(self) -> List[Any]:
        return self.val


from .utils.collector import Collector, ICollector
from .utils.class_util import ensure_subset
from .utils.dataloader import build_list, build_object, copy_dataloader_kls
from .utils.conversion import mapper
from .exceptions import (
    ResolverTargetAttrNotFound,
    LoaderFieldNotProvidedError,
    MissingAnnotationError,
    GlobalLoaderFieldOverlappedError,
    MissingCollector)
from .resolver import Resolver
from .utils.depend import LoaderDepend, Loader
from .utils.openapi import (
    model_config)


__all__ = [
    'Resolver',
    'LoaderDepend',
    'Loader',  # short 
    'Collector',
    'ICollector',

    'ResolverTargetAttrNotFound',
    'LoaderFieldNotProvidedError',
    'MissingAnnotationError',
    'GlobalLoaderFieldOverlappedError',
    'MissingCollector',

    'build_list',
    'build_object',
    'mapper',
    'ensure_subset',
    'model_config',
    'copy_dataloader_kls',
]


class ResolverTargetAttrNotFound(Exception):
    pass

class LoaderFieldNotProvidedError(Exception):
    pass

class MissingAnnotationError(Exception):
    pass

class GlobalLoaderFieldOverlappedError(Exception):
    pass

class MissingCollector(Exception):
    pass